#include "CThreadPool.h"

CThreadPool::CThreadPool(int thread_count) :
	mThreadCount(thread_count),
	mThreads{},
	mThreadID{},
	mJobs{},
	mbStopAll(false),
	mRequestCnt(0),
	mActiveJobCnt(0)
{
	InitializeCriticalSection(&mlock);
	int ThreadNo = 0;
	for (int i = 0; i < mThreadCount; ++i)
	{
		mThreads.push_back(CreateThread(NULL, NULL, &CThreadPool::WorkProcess, this, NULL, (LPDWORD)&ThreadNo));
		mThreadID.push_back(ThreadNo);
	}
}

CThreadPool::~CThreadPool()
{
	mbStopAll = true;

	WakeAllConditionVariable(&mcondition_v);
	DWORD result = WaitForMultipleObjects(mThreadCount, mThreads.data(), TRUE, INFINITE);

	for (int i = 0; i < mThreadCount; ++i)
	{
		CloseHandle(mThreads[i]);
		mThreads[i] = nullptr;
	}

	mThreadID.clear();
	DeleteCriticalSection(&mlock);
}

void CThreadPool::Release()
{

}

DWORD __stdcall CThreadPool::WorkProcess(LPVOID lpParam)
{
	CThreadPool* pool = reinterpret_cast<CThreadPool*>(lpParam);

	while (1)
	{
		EnterCriticalSection(&pool->mlock);

		if (pool->mJobs.empty() == true)
		{
			SleepConditionVariableCS(&pool->mcondition_v, &pool->mlock, INFINITE);
		}
		

		if (pool->mJobs.empty() == true)
		{
			LeaveCriticalSection(&pool->mlock);
			if (pool->mbStopAll == true)
				return 0;
			else
				continue;
		}

		if (pool->mbStopAll == true)
		{
			//thread 
			LeaveCriticalSection(&pool->mlock);
			return 0;
		}

		std::function<void()> work = std::move(pool->mJobs.front());
		pool->mJobs.pop();
		LeaveCriticalSection(&pool->mlock);
		
		work();
	}
	return 0;
}

bool CThreadPool::PushJob(std::function<void()> job)
{
	if (mbStopAll)
	{
		std::cout << "ThreadPool ." << std::endl;
		return false;
	}

	EnterCriticalSection(&mlock);
	mJobs.push(std::move(job));
	LeaveCriticalSection(&mlock);

	WakeConditionVariable(&mcondition_v);
	return true;
}

void CThreadPool::AlertableJob()
{
	WakeConditionVariable(&mcondition_v);
}




